/*
ConferenceParams.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;
import java.lang.StackTraceElement;
import org.linphone.mediastream.Log;

/**
* Object defining parameters for a {@link Conference}. <br/>
* <br/>
* Can be created by calling function {@link Core#createConferenceParams}. <br/>
*/
public interface ConferenceParams {
    /**
    * Enable audio capabilities. <br/>
    * <br/>
    * @param enable If true, audio will be enabled during conference <br/>
    */
    public void setAudioEnabled(boolean enable);

    /**
    * Enable chat capabilities. <br/>
    * <br/>
    * @param enable If true, chat is enabled during conference <br/>
    */
    public void setChatEnabled(boolean enable);

    /**
    * Check whether audio capabilities are enabled. <br/>
    * <br/>
    * @return true if the conference supports audio capabilities, false otherwise <br/>
    */
    public boolean isAudioEnabled();

    /**
    * Check whether chat capabilities are enabled. <br/>
    * <br/>
    * @return true if the conference supports chat capabilities, false otherwise <br/>
    */
    public boolean isChatEnabled();

    /**
    * Returns whether local participant has to enter the conference. <br/>
    * <br/>
    * @return true if local participant is by default part of the conference, false<br/>
    * otherwise <br/>
    */
    public boolean isLocalParticipantEnabled();

    /**
    * Returns whether conference can have only one participant. <br/>
    * <br/>
    * @return true if the conference can have only one participant, false otherwise <br/>
    */
    public boolean isOneParticipantConferenceEnabled();

    /**
    * Check whether video capabilities are enabled. <br/>
    * <br/>
    * @return true if the conference supports video capabilities, false otherwise <br/>
    */
    public boolean isVideoEnabled();

    /**
    * Enable local participant to enter the conference. <br/>
    * <br/>
    * The local participant is the one driving the local {@link Core}. It uses the<br/>
    * local sound devices. The default value is true. Setting to false is mostly<br/>
    * helpful when using liblinphone on a server application. <br/>
    * @param enable true if local participant is automatically added to the<br/>
    * conference, false otherwise <br/>
    */
    public void setLocalParticipantEnabled(boolean enable);

    /**
    * Enable conference with one participant. <br/>
    * <br/>
    * @param enable true if conference can have only one participant, false otherwise <br/>
    */
    public void setOneParticipantConferenceEnabled(boolean enable);

    /**
    * Enable video capabilities. <br/>
    * <br/>
    * @param enable If true, video will be enabled during conference <br/>
    */
    public void setVideoEnabled(boolean enable);

    /**
    * Clone a {@link ConferenceParams}. <br/>
    * <br/>
    * @return An allocated {@link ConferenceParams} with the same parameters than<br/>
    * params   <br/>
    */
    @NonNull
    public ConferenceParams clone();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();

    /**
      * Gets the native pointer used by this class to make native method calls.
      * @return the nativer pointer, as long
      */
    public long getNativePointer();

    public String toString();
}

class ConferenceParamsImpl implements ConferenceParams {

    protected long nativePtr = 0;
    protected transient Object userData = null;
    protected boolean _isConst = false;

    protected ConferenceParamsImpl(long ptr, boolean isConst) {
        nativePtr = ptr;
        _isConst = isConst;
    }

    public long getNativePointer() {
        return nativePtr;
    }


    public boolean isConst() {
        return _isConst;
    }

    private native void setAudioEnabled(long nativePtr, boolean enable);
    @Override
    synchronized public void setAudioEnabled(boolean enable)  {
        
        if (_isConst) {
			try {
				throw new CoreException(this.toString() + " is const! If you want to call setAudioEnabled() on it, clone it first.");
			} catch (CoreException e) {
				Log.e(e);
				for (StackTraceElement st : e.getStackTrace()) {
					Log.e(st);
				}
			}
		}
        setAudioEnabled(nativePtr, enable);
    }

    private native void setChatEnabled(long nativePtr, boolean enable);
    @Override
    synchronized public void setChatEnabled(boolean enable)  {
        
        if (_isConst) {
			try {
				throw new CoreException(this.toString() + " is const! If you want to call setChatEnabled() on it, clone it first.");
			} catch (CoreException e) {
				Log.e(e);
				for (StackTraceElement st : e.getStackTrace()) {
					Log.e(st);
				}
			}
		}
        setChatEnabled(nativePtr, enable);
    }

    private native boolean isAudioEnabled(long nativePtr);
    @Override
    synchronized public boolean isAudioEnabled()  {
        
        
        return isAudioEnabled(nativePtr);
    }

    private native boolean isChatEnabled(long nativePtr);
    @Override
    synchronized public boolean isChatEnabled()  {
        
        
        return isChatEnabled(nativePtr);
    }

    private native boolean isLocalParticipantEnabled(long nativePtr);
    @Override
    synchronized public boolean isLocalParticipantEnabled()  {
        
        
        return isLocalParticipantEnabled(nativePtr);
    }

    private native boolean isOneParticipantConferenceEnabled(long nativePtr);
    @Override
    synchronized public boolean isOneParticipantConferenceEnabled()  {
        
        
        return isOneParticipantConferenceEnabled(nativePtr);
    }

    private native boolean isVideoEnabled(long nativePtr);
    @Override
    synchronized public boolean isVideoEnabled()  {
        
        
        return isVideoEnabled(nativePtr);
    }

    private native void setLocalParticipantEnabled(long nativePtr, boolean enable);
    @Override
    synchronized public void setLocalParticipantEnabled(boolean enable)  {
        
        if (_isConst) {
			try {
				throw new CoreException(this.toString() + " is const! If you want to call setLocalParticipantEnabled() on it, clone it first.");
			} catch (CoreException e) {
				Log.e(e);
				for (StackTraceElement st : e.getStackTrace()) {
					Log.e(st);
				}
			}
		}
        setLocalParticipantEnabled(nativePtr, enable);
    }

    private native void setOneParticipantConferenceEnabled(long nativePtr, boolean enable);
    @Override
    synchronized public void setOneParticipantConferenceEnabled(boolean enable)  {
        
        if (_isConst) {
			try {
				throw new CoreException(this.toString() + " is const! If you want to call setOneParticipantConferenceEnabled() on it, clone it first.");
			} catch (CoreException e) {
				Log.e(e);
				for (StackTraceElement st : e.getStackTrace()) {
					Log.e(st);
				}
			}
		}
        setOneParticipantConferenceEnabled(nativePtr, enable);
    }

    private native void setVideoEnabled(long nativePtr, boolean enable);
    @Override
    synchronized public void setVideoEnabled(boolean enable)  {
        
        if (_isConst) {
			try {
				throw new CoreException(this.toString() + " is const! If you want to call setVideoEnabled() on it, clone it first.");
			} catch (CoreException e) {
				Log.e(e);
				for (StackTraceElement st : e.getStackTrace()) {
					Log.e(st);
				}
			}
		}
        setVideoEnabled(nativePtr, enable);
    }

    private native ConferenceParams clone(long nativePtr);
    @Override @NonNull
    synchronized public ConferenceParams clone()  {
        
        
        return (ConferenceParams)clone(nativePtr);
    }

    private native boolean unref(long ptr, boolean isConst);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
        if (nativePtr != 0) {
            boolean destroyed = unref(nativePtr, _isConst);
            if (destroyed) {
                nativePtr = 0;
            }
        }
        super.finalize();
    }

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }

    @Override
    public String toString() {
        return "Java object [" + super.toString() + "], native pointer [" + String.format("0x%08x", nativePtr) + "]";
    }
}
